//
// core.c
//

#include <assert.h>
#include "core.h"


/**
 * Memory locations for virtual X, Y, Z, T, A, B registers.
 */
  char reg[REG_NR][NUMBER_SIZE];

/**
 * Holds flags (INFTY, DIV0...).
 *
 * This is a special "flags" byte:
 * Possible bits so far:
 *
 * [  7       6       5       4       3       2       1       0  ]
 *                                                          INFTY
 */
  char flags;


/**
 * Copy contents of src to dst, whose contents will be lost.
 */
void copy_reg(char* dst, char* src) {
  memcpy(dst, src, NUMBER_SIZE);
}

/**
 * Return exponent of register.
 */
int get_e(char* r) {
  int e;

  e = r[E_TENS]*10 + r[E_ONES];
  if (r[SIGNS] & NEG_E) e = -e;
  return e;
}


/**
 * Increase exponent by 1.
 */
void inc_e(char* r) {

  if (r[SIGNS] & NEG_E) {                       // exp sign -
    if ((r[E_TENS] == 0) && (r[E_ONES] == 1)) { // exp = -1
      r[SIGNS] ^= NEG_E;                        // flip sign, is + now
      r[E_TENS] = 0;                            // exp = +0
      r[E_ONES] = 0;
    } else {                                    // exp < -1
      r[E_ONES]--;                              // just decreasing
      if (r[E_ONES] == -1) {                    // but watch for underflows
        r[E_TENS]--;
        r[E_ONES] = 9;
      }
    }
  } else {                                      // exp sign +
    if ((r[E_TENS] == 9) && (r[E_ONES] == 9)) { // exp = +99
      memset(&r[1], 0, NUMBER_SIZE-1);           // clear r, save SIGNS
      flags |= INFTY;                           // set INFTY flag
      debug_print("inc_e resulted in INFTY.");
    } else {                                    // exp < +99
      r[E_ONES]++;                              // just increasing
      if (r[E_ONES] == 10) {                    // but watch for overflows
        r[E_ONES] = 0;
        r[E_TENS]++;
      }
    }
  }
}

/**
 * Shift mantissa 1 digit to the right.
 *
 * Last digit is lost, but correct rounding on second-to-last digit
 * is assured.
 */
void shiftr1(char* r) {
  inc_e(r);
  if (r[M_ONES] >= 5) r[M_ONES-1]++;      // correct rounding
  int i;
  for (i=M_ONES; i>=M_HIGH; i--) r[i] = r[i-1];
  r[M_HIGH] = 0;
}

/**
 * Shift mantissa n digits to the right.
 */
void shiftr(char* r, int n) {
  { assert(n > 0); }
  for (; n > 0; n--) shiftr1(r);
}


/**
 * Add two registers and store result in third one.
 *
 * res <-- r1 + r2.
 * Even in case of overflow, res is changed in the process.
 */
void plus(char* res, char* r1, char* r2) {
  memset(res, 0, NUMBER_SIZE);            // result is cleared first
  int e1;
  int e2;

  e1 = get_e(r1);
  e2 = get_e(r2);

  if (e1 > e2) shiftr(r2, e1-e2);
  else if (e2 > e1) shiftr(r1, e2-e1);

  // Ok, exponents equalized. Now do actual addition of mantissa.

  res[E_TENS] = r1[E_TENS];   // use same exponent
  res[E_ONES] = r1[E_ONES];

  if ((!r1[SIGNS] & NEG_M) && (!r2[SIGNS] & NEG_M)) {   // both positive
    int i;
    for (i=M_ONES; i>M_HIGH; i--) {
      res[i] += r1[i] + r2[i];
      if (res[i] >= 10) {                 // overflow
        res[i-1]++;
        res[i] -= 10;
      }
    }
    res[M_HIGH] += r1[M_HIGH] + r2[M_HIGH];
    if (res[M_HIGH] >= 10) {
      res[M_HIGH] -= 10;
      shiftr1(res);
      if (flags) return;
      res[M_HIGH]++;
    }
  } else
  if ((r1[SIGNS] & NEG_M) && (r2[SIGNS] & NEG_M)) {  // both negative
    r1[SIGNS] ^= NEG_M;   // r1 = -r1
    r2[SIGNS] ^= NEG_M;   // r2 = -r2
    plus(res, r1, r2);                // TODO watch out for overflow here!
    res[SIGNS] ^= NEG_M;  // res = -res
  } else
  if ((r1[SIGNS] & NEG_M) ^ (r2[SIGNS] & NEG_M)) {   // one neg, one pos
    if (r1[SIGNS] & NEG_M) {
      r1[SIGNS] ^= NEG_M;   // r1 = -r1
      r2[SIGNS] ^= NEG_M;   // r2 = -r2
    }
//    sub(res, r1, r2);
  }
}


/**
 * Deletes top of stack.
 */
void pop() {
  copy_reg(reg[X], reg[Y]);
  copy_reg(reg[Y], reg[Z]);
  copy_reg(reg[Z], reg[T]);
}


/**
 * Put argument into X register and shifts the stack.
 *
 * Argument should be one of reg[A] or reg[B].
 */
void push(char* r) {
  copy_reg(reg[T], reg[Z]);
  copy_reg(reg[Z], reg[Y]);
  copy_reg(reg[Y], reg[X]);
  copy_reg(reg[X], r);
}


/**
 * Perform stack function "enter".
 */
void enter() {
  copy_reg(reg[T], reg[Z]);
  copy_reg(reg[Z], reg[Y]);
  copy_reg(reg[Y], reg[X]);
}


void init_core() {
  //example
  reg[T][SIGNS]  = NEG_M | NEG_E;
  reg[T][M_HIGH] = 3;
  reg[T][M+1]    = 1;
  reg[T][M+2]    = 4;
  reg[T][M+3]    = 1;
  reg[T][M+4]    = 5;
  reg[T][M+5]    = 9;
  reg[T][M+6]    = 2;
  reg[T][M+7]    = 6;
  reg[T][M+8]    = 5;
  reg[T][M_ONES] = 4;
  reg[T][E_TENS] = 0;
  reg[T][E_ONES] = 0;

  reg[Z][SIGNS]  = NEG_M;
  reg[Z][M_HIGH] = 5;
  reg[Z][M+1]    = 2;
  reg[Z][M+2]    = 6;
  reg[Z][M+3]    = 7;
  reg[Z][M+4]    = 3;
  reg[Z][M+5]    = 7;
  reg[Z][M+6]    = 2;
  reg[Z][M+7]    = 3;
  reg[Z][M+8]    = 6;
  reg[Z][M_ONES] = 7;
  reg[Z][E_TENS] = 2;
  reg[Z][E_ONES] = 0;

  reg[Y][SIGNS]  = 0x00;
  reg[Y][M_HIGH] = 9;
  reg[Y][M+1]    = 8;
  reg[Y][M+2]    = 2;
  reg[Y][M+3]    = 3;
  reg[Y][M+4]    = 4;
  reg[Y][M+5]    = 6;
  reg[Y][M+6]    = 3;
  reg[Y][M+7]    = 7;
  reg[Y][M+8]    = 5;
  reg[Y][M_ONES] = 3;
  reg[Y][E_TENS] = 9;
  reg[Y][E_ONES] = 9;

  reg[X][SIGNS]  = 0x00;
  reg[X][M_HIGH] = 7;
  reg[X][M+1]    = 4;
  reg[X][M+2]    = 4;
  reg[X][M+3]    = 7;
  reg[X][M+4]    = 3;
  reg[X][M+5]    = 3;
  reg[X][M+6]    = 7;
  reg[X][M+7]    = 3;
  reg[X][M+8]    = 4;
  reg[X][M_ONES] = 7;
  reg[X][E_TENS] = 9;
  reg[X][E_ONES] = 9;
} // init_regs
